//Copyright 2012-2013 Haley Abel
//
//This file is part of ClusterFAST.
//
//ClusterFAST is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//
//ClusterFAST is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License along with ClusterFAST.  If not, see <http://www.gnu.org/licenses/>.
import java.util.ArrayList;
import java.util.List;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Vector;
import java.util.Iterator;
import java.util.Collections;

import net.sf.samtools.SAMRecord;
import net.sf.samtools.SAMRecordIterator;



public class ClusterMap {

    private ArrayList< LinkedHashMap<String, LinkedHashMap<Integer, ArrayList<SAMRecord> > > > clusmap=null;
    private int stepsize=0;
    private int tol=0;
    private int minct=2;
    
    public ClusterMap() {
    }
    
    public ClusterMap(ArrayList<ArrayList<SAMRecord> >sams, ArrayList< LinkedHashMap<String, LinkedHashSet<Integer> > > samrecs, int ss, int tl, int min, ArrayList< LinkedHashMap<String, LinkedHashSet<String> > > materecs) {
        
        tol=tl;
        stepsize=ss;
        minct=min;
        clusmap=new ArrayList<LinkedHashMap<String, LinkedHashMap<Integer, ArrayList<SAMRecord> > > >();
		
		for(int j=0; j<sams.size(); j++) {
          
			clusmap.add(new LinkedHashMap<String, LinkedHashMap<Integer, ArrayList<SAMRecord> > >());
			LinkedHashMap<String, LinkedHashSet<Integer> > srj=samrecs.get(j);
        
			for(String name:srj.keySet()) {
				LinkedHashMap<Integer, Integer> cts=new LinkedHashMap<Integer, Integer>();
				ArrayList<Integer> recs=new ArrayList<Integer> (srj.get(name));
            
				for(int i=0; i<recs.size(); i++) {
			
					double f100=((double)recs.get(i))/stepsize;
					int pos100=Math.round((float)f100);
                    if(!cts.containsKey(pos100)) {
						cts.put(pos100, 0);
						int floor=(int) Math.floor(f100);
						int ceil=(int) Math.ceil(f100);
						if(pos100==floor && cts.containsKey(ceil)) 
							cts.put(ceil, cts.get(ceil)+1);
						else if (pos100==ceil && cts.containsKey(floor)) 
							cts.put(floor, cts.get(floor)+1);
					}	
					cts.put(pos100, cts.get(pos100)+1);
				}
				for(Integer i: cts.keySet()) {
					if(cts.get(i)>=minct) {
						if(!clusmap.get(j).containsKey(name)) 
							clusmap.get(j).put(name, new LinkedHashMap<Integer, ArrayList<SAMRecord> >());
						if(!clusmap.get(j).get(name).containsKey(i)) 
						    clusmap.get(j).get(name).put(i, new ArrayList<SAMRecord>());
					      
					}
				}
			}
			
		}
        this.fill(sams, materecs);
    }
    
    public void fill(ArrayList<ArrayList<SAMRecord> >sams, ArrayList< LinkedHashMap<String, LinkedHashSet<String> > > materecs) {
	
		for(int j=0; j<sams.size(); j++) {
			System.out.println(j);
			ArrayList<SAMRecord> sj=sams.get(j);
            LinkedHashMap<String, LinkedHashSet<String> > mates=materecs.get(j);
        
			for(int i=0; i<sj.size(); i++) {
				SAMRecord rec=sj.get(i);
				String name=rec.getMateReferenceName();
				 
				if(clusmap.get(j).containsKey(name)) {
					
					float f100=((float)rec.getMateAlignmentStart())/stepsize;
					int pos100=Math.round(f100);
                    
                    if(clusmap.get(j).get(name).containsKey(pos100)) {
                        if(mates.containsKey(name) && mates.get(name).contains(rec.getReadName())) 
                            clusmap.get(j).get(name).get(pos100).add(rec);
					}
					else {
						for(int h=0; h<tol; h++) {
							int floor=(int)(Math.floor(f100-h));
							int ceil=(int)(Math.ceil(f100+h));
                        
							double dist=99;
							int best=-99;
							if(clusmap.get(j).get(name).containsKey(ceil)) {
								dist=ceil-f100;
								best=ceil;
							}
							if(clusmap.get(j).get(name).containsKey(floor)) {
								if(f100-floor<dist)
									best=floor;
							}
							if(best>-1) {
                                if(mates.containsKey(name) && mates.get(name).contains(rec.getReadName())) 
                                    clusmap.get(j).get(name).get(best).add(rec);
								break;
							}
                        }
                    }
                }
            }
        } 
    }
    
    public int getTol() {
        return tol;
    }
    
    public int getStepsize() {
        return stepsize;
    }
    
    public void print() {
		for(int j=0; j<clusmap.size(); j++) {
			for(String name: clusmap.get(j).keySet()) {
				for(Integer k: clusmap.get(j).get(name).keySet()) 
					System.out.println(j+"\t"+name+"\t"+k+"\t"+clusmap.get(j).get(name).get(k).size());
            }
        }
    }
	
	public ArrayList< LinkedHashMap<String, LinkedHashMap<Integer, ArrayList<SAMRecord> > > > getMap() {
		return clusmap;
	}
        
}
