//Copyright 2012-2013 Haley Abel
//
//This file is part of ClusterFAST.
//
//ClusterFAST is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//
//ClusterFAST is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License along with ClusterFAST.  If not, see <http://www.gnu.org/licenses/>.
import java.io.File;
import java.io.*;
import java.util.ArrayList;
import java.util.List;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Collections;
import java.util.Comparator;

import net.sf.samtools.SAMFileReader;
import net.sf.samtools.SAMFileWriter;
import net.sf.samtools.SAMFileWriterFactory;
import net.sf.samtools.SAMRecord;
import net.sf.samtools.SAMRecordIterator;
import net.sf.samtools.SAMFileReader.ValidationStringency;

public class QPosMap {

    private ArrayList<LinkedHashMap<String, LinkedHashMap<Integer, LinkedHashMap<Integer, LinkedHashMap<String, Integer> > > > > pairs=null;

	private ArrayList< LinkedHashMap<String, ArrayList<Integer> > > qposmap=null;
    private ArrayList<Interval1> intervals=null;
	int stepsize=-1;
	int minct=-1;
    int tol=-1;
	
	
    public QPosMap(ArrayList<Interval1> ints, int ss) {
        intervals=ints;
		stepsize=ss;
		qposmap=new ArrayList< LinkedHashMap<String, ArrayList<Integer> > >();
		for(int j=0; j<intervals.size(); j++) {
			qposmap.add(new LinkedHashMap<String, ArrayList<Integer> >());
		}
	}
	
	public void readPairs(String qposfile) {
	
	    pairs=new ArrayList<LinkedHashMap<String, LinkedHashMap<Integer,  LinkedHashMap<Integer, LinkedHashMap<String, Integer> >  > > >();  
		for(int k=0; k<intervals.size(); k++) 
		    pairs.add(new LinkedHashMap<String, LinkedHashMap<Integer,  LinkedHashMap<Integer, LinkedHashMap<String, Integer> > > > ());  
		try {				
			BufferedReader input =  new BufferedReader(new FileReader(qposfile));			//change to qposfileold
			String line=null;
			while (( line = input.readLine()) != null){
					
				String[] spl=line.split("\t");
					
				int itxnum=Integer.parseInt(spl[0]);
				int itxpos=0;
				int nonitxpos=0;
				if(spl[2].length()>2) 
					nonitxpos=Integer.parseInt(spl[2].substring( 0, spl[2].length()-2));
				if(spl[4].length()>2)
					itxpos=Integer.parseInt(spl[4].substring( 0, spl[4].length()-2));
				itxpos=(itxpos*100+(itxpos+1)*100)/(2*stepsize);
				nonitxpos=(nonitxpos*100+(nonitxpos+1)*100)/(stepsize*2);
			
				String orientation=spl[5];
				int supporting_reads=Integer.parseInt(spl[6]);
				
				if(!pairs.get(itxnum).containsKey(spl[1])) 
				    pairs.get(itxnum).put(spl[1], new LinkedHashMap< Integer,  LinkedHashMap<Integer, LinkedHashMap< String, Integer> > >()); 
				   
				if(!pairs.get(itxnum).get(spl[1]).containsKey(nonitxpos))
                    pairs.get(itxnum).get(spl[1]).put(nonitxpos, new  LinkedHashMap<Integer, LinkedHashMap< String, Integer> >());
				if(!pairs.get(itxnum).get(spl[1]).get(nonitxpos).containsKey(itxpos)) 
				    pairs.get(itxnum).get(spl[1]).get(nonitxpos).put(itxpos, new LinkedHashMap<String, Integer>()); 
				   
				pairs.get(itxnum).get(spl[1]).get(nonitxpos).get(itxpos).put(orientation,supporting_reads);
			}
			input.close();
		}
		catch(Exception e) {
		}
	}
	
	
	public void fill(ClusterMap cmap, boolean filter, String qposfile, int minct1) {
	
		minct=minct1;
        int pairtol=250;
        tol=cmap.getTol();
	ArrayList<Entry> entries=new ArrayList<Entry>();
		try {
		
			PrintStream ps=new PrintStream(new File(qposfile));
		
			for(int k=0; k<intervals.size(); k++) {
		
				LinkedHashMap<String, LinkedHashMap<Integer, ArrayList<SAMRecord> > > clustermap=cmap.getMap().get(k);
				String itxname=intervals.get(k).getName();
			
				for(String name :clustermap.keySet()) {
					for(Integer j: clustermap.get(name).keySet()) {
						ArrayList<SAMRecord> samv=clustermap.get(name).get(j);
                        LinkedHashMap<Integer, LinkedHashSet<String> > uniquepos_by_flag=new  LinkedHashMap<Integer, LinkedHashSet<String> >();
						for(int i=0; i<samv.size(); i++) {
							
							int astart=samv.get(i).getAlignmentStart();
                            String mateneg=samv.get(i).getMateNegativeStrandFlag() ? "-":"+";
							String readneg=samv.get(i).getReadNegativeStrandFlag() ? "-":"+";
							String comb=mateneg+readneg;
							if(!uniquepos_by_flag.containsKey(astart)) 
							    uniquepos_by_flag.put(astart, new LinkedHashSet<String>());
							uniquepos_by_flag.get(astart).add(comb);
						}
						
												//always true
					   if(uniquepos_by_flag.keySet().size()>=1) {
							
                            LinkedHashMap<Integer, LinkedHashMap<String, Integer> > uniquemap_by_flag=new LinkedHashMap<Integer, LinkedHashMap<String, Integer> >();
							for(Integer i: uniquepos_by_flag.keySet() ) {
								double f100=(double) i/stepsize;
							    int pos100=Math.round((float)f100);
								int ceil=(int) Math.ceil(f100);
								int floor=(int) Math.floor(f100);

								if(uniquemap_by_flag.get(pos100)==null) 
								    uniquemap_by_flag.put(pos100, new LinkedHashMap<String, Integer>());
								   
								for(String h : uniquepos_by_flag.get(i)) {
									if(uniquemap_by_flag.get(pos100).get(h)==null) {
										uniquemap_by_flag.get(pos100).put(h, 0);
										
										if(ceil==pos100 && uniquemap_by_flag.get(floor)!=null && uniquemap_by_flag.get(floor).containsKey(h))
											uniquemap_by_flag.get(floor).put(h, uniquemap_by_flag.get(floor).get(h)+1);
										else if(floor==pos100 && uniquemap_by_flag.get(ceil)!=null && uniquemap_by_flag.get(ceil).containsKey(h))
											uniquemap_by_flag.get(ceil).put(h, uniquemap_by_flag.get(ceil).get(h)+1);

									}
									uniquemap_by_flag.get(pos100).put(h, uniquemap_by_flag.get(pos100).get(h)+1);
								}
								
							}
						
							for(Integer i: uniquemap_by_flag.keySet()) {
								boolean prnt=false;
								for(String h: uniquemap_by_flag.get(i).keySet()) {
								    if(uniquemap_by_flag.get(i).get(h)>=minct) 
										prnt=true;
										
								}
								if(prnt) {
									for (String h:uniquemap_by_flag.get(i).keySet()) {
									String str=k+"\t"+name+"\t"+j*stepsize+"\t"+itxname+"\t"+i*stepsize+"\t"+h+"\t"+uniquemap_by_flag.get(i).get(h)+"\t";
									boolean pass=false;
									int max_supp=0;
									if(!filter) {
										pass=true;
										System.out.println(str);
										ps.println(str);
									}
									else {		//filter by pairs
										LinkedHashMap<String, LinkedHashMap<Integer,  LinkedHashMap<Integer, LinkedHashMap<String, Integer> > > > pfilter=pairs.get(k);
										
										if(pfilter.containsKey(name) ) {
											for(int kk=-1*pairtol; kk<(pairtol+1); kk++) {
												if(pfilter.get(name).containsKey(j+kk)) {
													for(int jj=-1*pairtol; jj<(pairtol+1); jj++) {
														if(pfilter.get(name).get(j+kk).containsKey(i+jj)) {
															if(pfilter.get(name).get(j+kk).get(i+jj).containsKey(h)) {
																if (pfilter.get(name).get(j+kk).get(i+jj).get(h)>0) {
																	pass=true;
																	entries.add(new Entry(k, itxname, (i+jj)*stepsize, name, (j+kk)*stepsize, uniquemap_by_flag.get(i).get(h), h, pfilter.get(name).get(j+kk).get(i+jj).get(h)));
																}
															}
														}
													}
												}
											}
										}
									}
									if(pass) {
                                    
										if(!qposmap.get(k).keySet().contains(name))
											qposmap.get(k).put(name, new ArrayList<Integer>());
										qposmap.get(k).get(name).add(j);
										if(!qposmap.get(k).keySet().contains(itxname))
											qposmap.get(k).put(itxname, new ArrayList<Integer>());
										qposmap.get(k).get(itxname).add(i);
									}
								}
							}
						}
						}					
					}
				}
			}
			
			Collections.sort(entries, new Entry.CompEntry());
        
			if (filter) {
			    Entry.CompSome cs=new Entry.CompSome();
                Entry oldentry=null;
			   // Entry oldentry=entries.get(0);
			    LinkedHashSet<Entry> toRemove=new LinkedHashSet<Entry>();
			    for (int h=0; h<entries.size(); h++) {
                    if(oldentry!=null && cs.compare(oldentry, entries.get(h))==0) 
                        toRemove.add(entries.get(h));
                    else 
                        oldentry=entries.get(h);
			    }
			    entries.removeAll(toRemove);    

			    for(int h=0; h<entries.size(); h++) {
                    System.out.println(entries.get(h));
                    ps.println(entries.get(h));

			    }
			}
			ps.close();
		}
		catch(FileNotFoundException e) {		
            System.err.println("Error:  "+qposfile+" not found.\n");
		}
	}
	
	public  ArrayList< LinkedHashMap<String, ArrayList<Integer> > > getMap() {
		return qposmap;
	}
    public static class Entry {

	public int interval;
	public String itxname;
	public String partner;
	public int ppos;
	public int itxpos;
	public int count;
	public String orientation;
	public int paircount;
	public String outstr;

	public Entry(int intnum, String iname, int ipos, String pname, int pos, int ct, String or, int pcount) {
	    interval=intnum;
	    itxname=iname;
	    itxpos=ipos;
	    partner=pname;
	    ppos=pos;
	    count=ct;
	    orientation=or;
	    paircount=pcount;
	}
	public String toString() {
	    return interval+"\t"+partner+"\t"+ppos+"\t"+itxname+"\t"+itxpos+"\t"+orientation+"\t"+count+"\t"+paircount;
	}

	public static class CompEntry implements Comparator<Entry> {
	    @Override
		public int compare(Entry e0, Entry e1) {
		if(e0.interval!=e1.interval) {
		    return e0.interval-e1.interval;
		} else if (e0.itxname.compareTo(e1.itxname)!=0) {
		    return e0.itxname.compareTo(e1.itxname);
		} else if (e0.partner.compareTo(e1.partner)!=0) {
		    return e0.partner.compareTo(e1.partner);
		} else if (e0.orientation.compareTo(e1.orientation)!=0) {
		    return e0.orientation.compareTo(e1.orientation);
		} else if (e0.count!=e1.count) {
		    return e1.count-e0.count;
		} else if (e0.paircount!=e1.paircount) {
		    return (e1.paircount-e0.paircount);
		} else if (e0.itxpos!=e1.itxpos) {
		    return e0.itxpos-e1.itxpos;
		} else if (e0.ppos!=e1.ppos) {
		    return e0.ppos-e1.ppos;
		} else {
		    return 0;
		}
	    }
	}
	
	public static class CompSome implements Comparator<Entry> {
	    @Override
		public int compare(Entry e0, Entry e1) {
		if(e0.interval!=e1.interval) {
            return e0.interval-e1.interval;
        } else if (e0.itxname.compareTo(e1.itxname)!=0) {
            return e0.itxname.compareTo(e1.itxname);
        } else if (e0.partner.compareTo(e1.partner)!=0) {
            return e0.partner.compareTo(e1.partner);
        } else if (e0.orientation.compareTo(e1.orientation)!=0) {
            return e0.orientation.compareTo(e1.orientation);
		} else if (e0.itxpos-e1.itxpos>100) {
		    return 1;
		} else if (e1.itxpos-e0.itxpos>100) {
		    return -1;
		} else if (e0.ppos-e1.ppos>100) {
		    return 1;
		} else if (e1.ppos-e0.ppos>100) {
		    return -1;
		} else {
		    return 0;
		}
	    }
	}
    }
    
    public boolean matchesSingle(SAMRecord inrec, int xtol) {
        
        int pos100=inrec.getAlignmentStart()/stepsize;
        String name=inrec.getReferenceName();
        boolean match=false;
        
        for(int k=0; k<intervals.size(); k++) {
            if(qposmap.get(k).keySet().contains(name)) {
                ArrayList<Integer> qpv=qposmap.get(k).get(name);
                for(int i=0; i<qpv.size(); i++) {
                    if(Math.abs(pos100-qpv.get(i))<(tol+xtol)) {
                        match=true;
                        break;
                    }
                }
            }
        }
        return match; 
    }
	
	public boolean matchesPair(SAMRecord inrec, ArrayList<Interval1> ints, int xtol) {
        
        boolean match=false;
        
        for(int k=0; k<intervals.size(); k++) {
		
			if (inrec.getReferenceName().equals(ints.get(k).getName()) || inrec.getMateReferenceName().equals(ints.get(k).getName())) {
				
				LinkedHashMap< String, ArrayList<Integer> >  qpmap=qposmap.get(k);
				int recpos100=0;
				int matepos100=0;
				String name=null;
					
				if (inrec.getReferenceName().equals(ints.get(k).getName())) {
					recpos100=inrec.getAlignmentStart()/stepsize;
					matepos100=inrec.getMateAlignmentStart()/stepsize;
					name=inrec.getMateReferenceName();
				}
				else {
					matepos100=inrec.getAlignmentStart()/stepsize;
					recpos100=inrec.getMateAlignmentStart()/stepsize;
					name=inrec.getReferenceName();
				}
					
				if(qpmap.keySet().contains(name)) {
						
					for(int kk=0; kk<qpmap.get(name).size(); kk++) {
						if(Math.abs(qpmap.get(name).get(kk)-matepos100)<(tol+xtol)) {
						
							for(int jj=0; jj<qpmap.get(ints.get(k).getName()).size(); jj++) {
								if(Math.abs(qpmap.get(ints.get(k).getName()).get(jj)-recpos100)<(tol+xtol)) {
									match=true;
									break;
								}
							}
						}
					}
				}																							
			}
        }
        return match; 
    }

	
}
		
