//Copyright 2012-2013 Haley Abel
//
//This file is part of ClusterFAST.
//
//ClusterFAST is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//
//ClusterFAST is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License along with ClusterFAST.  If not, see <http://www.gnu.org/licenses/>.
import java.io.File;
import java.io.*;
import java.util.ArrayList;
import java.util.List;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Vector;
import java.util.Iterator;
import java.util.Collections;


import net.sf.samtools.Cigar;
import net.sf.samtools.CigarElement;
import net.sf.samtools.CigarOperator;
import net.sf.samtools.SAMFileReader;
import net.sf.samtools.SAMFileWriter;
import net.sf.samtools.SAMFileWriterFactory;
import net.sf.samtools.SAMRecord;
import net.sf.samtools.SAMRecordIterator;
import net.sf.samtools.SAMFileReader.ValidationStringency;
import net.sf.samtools.*;

import net.sf.picard.reference.FastaSequenceFile;
import net.sf.picard.reference.ReferenceSequenceFileWalker;
import net.sf.picard.reference.ReferenceSequence;
import net.sf.picard.io.IoUtil;

import net.sf.picard.filter.IntervalFilter;
import net.sf.picard.util.*;
import net.sf.picard.sam.*;


public class TxFinder_2 {
    
    private SAMFileReader inputSam=null;
    private String outdir=null;
    private String itxfile=null;
    private String qposfile=null;
	private String qposfileold=null;
	private boolean firstPass=true;
    
    private String samoutDiscordant=null;
    private String samoutSingle=null;
    private String samoutMateUnmapped=null;
    
    private String samoutDiscordantFiltered=null;
    private String samoutSingleFiltered=null;
    private String samoutMateUnmappedFiltered=null;
    
    private String fq1=null;
    private String fq2=null;
    

    
    
 /*   private String samout2=null;
    private String samout1=null;
	private String samout3=null;
    private String samout1final=null;
	private String samout2final=null;
	private String samout3final=null;*/
   
  	     
	
	private ArrayList<LinkedHashMap<String, LinkedHashSet<Integer> > > samrecs=null;
    private ArrayList<LinkedHashMap<String, LinkedHashSet<String> > > materecs=null;
	private ArrayList<ArrayList<SAMRecord> > sams=null; 
    
    
    private ClusterMap clustermap=null;
    private QPosMap qposmap=null; 
    
    PrintStream ps=null;
    
    private ArrayList<Interval1> intervals=null;
    
  	
	private int minct1=1;
	private int distfromint=50000;
	
    
    public TxFinder_2() {
    }
	
    public TxFinder_2(SAMFileReader _inputSam, String _itxfile, int minct, int distance, String outdir1, boolean _firstPass) {
	  
        inputSam=_inputSam;
        inputSam.setValidationStringency(ValidationStringency.SILENT);
		firstPass=_firstPass;
		outdir=outdir1;
		distfromint=distance;
        itxfile=_itxfile;
		minct1=minct;
          
        readIntervals();
        makeDirectories();
		        
    }
    
    public void readIntervals() {
        
        ArrayList<Interval> intervals_tmp=new ArrayList<Interval>(IntervalList.fromFile(new File(itxfile)).getIntervals());
        intervals=new ArrayList<Interval1>();
        samrecs= new ArrayList <LinkedHashMap <String, LinkedHashSet<Integer> > >();
        materecs= new ArrayList <LinkedHashMap <String, LinkedHashSet<String> > >();
        sams=new ArrayList < ArrayList<SAMRecord> >();
        SAMFileHeader head=inputSam.getFileHeader();
        
        for(int i=0; i<intervals_tmp.size(); i++) {
			Interval _int=intervals_tmp.get(i);
			System.out.println(_int.getSequence()+"\t"+_int.getName()+"\t"+_int.getStart()+"\t"+_int.length());
            intervals.add(new Interval1(_int, head, distfromint));
            intervals.get(i).print();
            samrecs.add( new LinkedHashMap<String, LinkedHashSet<Integer> >());
            materecs.add( new LinkedHashMap<String, LinkedHashSet<String> >());
			sams.add( new ArrayList<SAMRecord>());
		}
    }
    
    public void makeDirectories() {
        
        System.out.println(outdir);
        String temp=outdir+"/temp";
        new File(temp).mkdirs();
        if(firstPass) {
            
            samoutMateUnmapped=temp+"/matenm.1.bam";			
			samoutDiscordant=temp+"/2ends.1.bam";
			samoutSingle=temp+"/1end.1.bam";
           
			samoutSingleFiltered=temp+"/1end.filtered.1.bam";
			samoutDiscordantFiltered=temp+"/2ends.filtered.1.bam";
			samoutMateUnmappedFiltered=temp+"/matenm.filtered.1.bam";	
          
			qposfile=outdir+"/breakpoints.1.txt";
            fq1=temp+"/toRemap1.fq";
            fq2=temp+"/toRemap2.fq";

            
			/*samout3=temp+"/matenm.1.sam";			
			samout2=temp+"/2ends.1.sam";
			samout1=temp+"/1end.1.sam";
            softclipssam=temp+"/softclips.1.sam";
			samout1final=temp+"/final.1.sam";
			samout2final=temp+"/2ends.final.1.sam";
			samout3final=temp+"/matenm.final.1.sam";	
            softclipsfinal=temp+"/softclips.final.1.sam";
			qposfile=outdir+"/breakpoints.1.txt";*/
		}
		else {
			System.out.println("pass 2");
            
            samoutMateUnmapped=temp+"/matenm.2.bam";		
			samoutDiscordant=temp+"/2ends.2.bam";
			samoutSingle=temp+"/1end.2.bam";
			samoutDiscordantFiltered=temp+"/filtered.2.bam";
			qposfile=outdir+"/breakpoints.2.txt";
			samoutMateUnmappedFiltered=temp+"/matenm.filtered.2.bam";	
			qposfileold=outdir+"/breakpoints.1.txt";
            
			/*samout3=temp+"/matenm.2.sam";		
			samout2=temp+"/2ends.2.sam";
			samout1=temp+"/1end.2.sam";
			samout1final=temp+"/final.2.sam";
			qposfile=outdir+"/breakpoints.2.txt";
			samout3final=temp+"/matenm.final.2.sam";	
			qposfileold=outdir+"/breakpoints.1.txt";	*/
		}

    }

    
    public void sortInputSam() {
       
        
        SAMFileWriterFactory sf1=new SAMFileWriterFactory();
        SAMFileWriter sw_single=sf1.makeSAMOrBAMWriter(inputSam.getFileHeader(), false, new File(samoutSingle));
       /* sf1.setCreateIndex(false);
        SAMFileWriter sw_single=sf1.makeSAMWriter(inputSam.getFileHeader(),false,new File(samoutSingle));*/
        
        SAMFileWriterFactory sf2=new SAMFileWriterFactory();
        SAMFileWriter sw_pair=sf2.makeSAMOrBAMWriter(inputSam.getFileHeader(),false,new File(samoutDiscordant));
        /*sf2.setCreateIndex(false);
        SAMFileWriter sw_pair=sf2.makeSAMWriter(inputSam.getFileHeader(),false,new File(samoutDiscordant));*/
		
		SAMFileWriterFactory sf3=null;
		SAMFileWriter sw_matenm=null;
	        		
		if(firstPass) {
			System.out.println(samoutMateUnmapped);
			sf3=new SAMFileWriterFactory();
            sw_matenm=sf3.makeSAMOrBAMWriter(inputSam.getFileHeader(),false,new File(samoutMateUnmapped));
            
        }
        
        SAMRecordIterator iter=inputSam.iterator();
        
        while(iter.hasNext())   {  
					
            SAMRecord rec=iter.next();
			if(rec.getDuplicateReadFlag())					//ignore duplicates
				continue;
           
			if (rec.getReadUnmappedFlag()) {               //reads that are unmapped but mate maps to interval, sam file funny here
				if(firstPass && !rec.getMateUnmappedFlag()) {
					for(int i=0; i<intervals.size(); i++) {
                        if(intervals.get(i).contains(rec.getReferenceIndex(), rec.getAlignmentStart())) 
                            sw_single.addAlignment(rec);
					}				
				}
			}
			else {
			
				if(rec.getMateUnmappedFlag()) {            //read maps to interval, mate is unmapped
                    if(firstPass) {
                        for(int i=0; i<intervals.size(); i++) {
                            if(intervals.get(i).contains(rec.getReferenceIndex(), rec.getAlignmentStart())) 
                                sw_matenm.addAlignment(rec);   
                        }	
                    }
				}
				else {                                      //both ends mapped
                    boolean scadded=false;
                    boolean swadded=false;
				
					for(int i=0; i<intervals.size(); i++) {
                    
						// if rec maps to interval
                        if(intervals.get(i).contains(rec.getReferenceIndex(), rec.getAlignmentStart())) {
							
							Cigar cig=rec.getCigar();
							if(cig.numCigarElements()>1) {
								List<CigarElement> els=cig.getCigarElements();
								for(CigarElement el : els) {
									// if rec doesn't map completely
									if(el.getOperator()==CigarOperator.SOFT_CLIP) {
										if(el.getLength()>10) {
											sw_single.addAlignment(rec);
                                            swadded=true;
                                        }
                                    }
								}
							}
							//and mate doesn't map to interval
                            boolean matenotint=false;
                            if (!(intervals.get(i).contains(rec.getMateReferenceIndex(), rec.getMateAlignmentStart() , 1000))) 
                                matenotint=true;
                            else if ((intervals.get(i).contains(rec.getMateReferenceIndex(), rec.getMateAlignmentStart() , 1000)) && Math.abs(rec.getMateAlignmentStart()-rec.getAlignmentStart())>1000)
                                matenotint=true;
                            if(matenotint) {
                            
								sw_pair.addAlignment(rec);
								String refname=rec.getMateReferenceName();
                                if(rec.getMappingQuality()>0||!firstPass) {
                                    LinkedHashSet<Integer> sriref=samrecs.get(i).get(refname);
                                    if(sriref!=null)
                                        sriref.add(rec.getMateAlignmentStart());
                                    else {
                                        samrecs.get(i).put(refname, new LinkedHashSet<Integer>());
                                        samrecs.get(i).get(refname).add(rec.getMateAlignmentStart());
                                    }
                                    sams.get(i).add(rec);
                                }
							}
						}
						// if mate maps to interval
					
                        if(intervals.get(i).contains(rec.getMateReferenceIndex(), rec.getMateAlignmentStart())) {
							
							Cigar cig=rec.getCigar();
							if(cig.numCigarElements()>1) {

								List<CigarElement> els=cig.getCigarElements();
								for(CigarElement el : els) {
									// but not completely
									if(el.getOperator()==CigarOperator.SOFT_CLIP) {
										if(el.getLength()>10) {
                                            if(!swadded)
                                                sw_single.addAlignment(rec);
                                        }
									}
								}
							}
							// and rec does not
                            boolean recnotint=false;
                            if (!(intervals.get(i).contains(rec.getReferenceIndex(), rec.getAlignmentStart() , 1000))) 
                                recnotint=true;
                            else if ((intervals.get(i).contains(rec.getReferenceIndex(), rec.getAlignmentStart() , 1000)) && Math.abs(rec.getAlignmentStart()-rec.getMateAlignmentStart())>1000)
                                recnotint=true;
                            
                            if(recnotint) {
                                sw_pair.addAlignment(rec);
								String refname=rec.getReferenceName();
                        
                                if(rec.getMappingQuality()>0||!firstPass) {
                                    LinkedHashSet<String> mriref=materecs.get(i).get(refname);
                                    if(mriref!=null) 
                                        mriref.add(rec.getReadName());
                                    else {
                                        materecs.get(i).put(refname, new LinkedHashSet<String>());
                                        materecs.get(i).get(refname).add(rec.getReadName());
                                    }
                                }
                            }
                            
						}
					}
				}
			}
           
		}
        iter.close();
        inputSam.close();
        sw_pair.close();
        sw_single.close();
		if(firstPass) {
			sw_matenm.close();
            //sw_sc.close();
		}
    }
    
    public void filterOtherSingleEnds() {
        
	filterSingleEnds(samoutMateUnmapped, samoutMateUnmappedFiltered, 2, false);
	}
    
    public void filterSingleEnds() {
        filterSingleEnds(false);
    }

    public void filterSingleEnds(boolean writeFa) {
	System.out.println("filter single ends.\n");
        filterSingleEnds(samoutSingle,  samoutSingleFiltered , 2, writeFa);
    }
    
   /* public void filterSoftClips() {
        System.out.println("filter soft clips\n");
        filterSingleEnds(softclipssam, softclipsfinal, 2, false);
    }*/

    
    public void filterSingleEnds(String infile, String outfile, int tolextra, boolean writeFa) {
        
        SAMFileReader tempSam=new SAMFileReader(new File(infile));
        tempSam.setValidationStringency(ValidationStringency.SILENT);
        SAMRecordIterator iter=tempSam.iterator();
        
        SAMFileWriterFactory sf2=new SAMFileWriterFactory();
        SAMFileWriter sw2=sf2.makeSAMOrBAMWriter(tempSam.getFileHeader(),false,new File(outfile));
       /* sf2.setCreateIndex(false);
        SAMFileWriter sw2=sf2.makeSAMWriter(tempSam.getFileHeader(),false,new File(outfile));*/
	
        try {
            BufferedWriter bw1=null;
            BufferedWriter bw2=null;
            if(writeFa) {
                bw1=new BufferedWriter(new FileWriter(new File (fq1)));
                bw2=new BufferedWriter(new FileWriter(new File (fq2)));
            }
        
            while(iter.hasNext()) {
                SAMRecord rec=iter.next();
                if(qposmap.matchesSingle(rec, tolextra)) { 
                    sw2.addAlignment(rec);
                    if(writeFa) {
                        StringBuilder f1=new StringBuilder();
                        StringBuilder f2=new StringBuilder();
                        if(splitSAM(rec, f1,  f2)==1) {
                            bw1.write(f1.toString());
                            bw2.write(f2.toString());
                        }
                    }
                }
            }
            sw2.close();	
            iter.close();
            tempSam.close();
            if(writeFa) {
                bw1.close();
                bw2.close();
            }
        } catch(IOException e) {
            System.err.println("Error in filterSingleEnds.\n");
        }
    }

    public int splitSAM(SAMRecord sam, StringBuilder f1, StringBuilder f2) {
        int len=40;
        int minlen=80;
        String read=sam.getReadString();
        if(read.length()<minlen)
            return 0;
        String quals=sam.getBaseQualityString();
        String qual1=quals.substring(0, Math.min(len, quals.length()));
        String qual2=quals.substring(Math.max(0, quals.length()-len));
        String read1=read.substring(0, Math.min(len, read.length()));
        String read2=read.substring(Math.max(0, read.length()-len));
        char[] read2c=new char[read2.length()];
        for(int i=0; i<read2c.length;i++) {
            switch(read2.charAt(i)) {
                case 'A': read2c[i]='T'; break;
                case 'T': read2c[i]='A'; break;
                case 'C': read2c[i]='G'; break;
                case 'G': read2c[i]='C'; break;
	    default:  System.err.println("Invalid base.\n"+read2.charAt(i)); return 0;
            }
        }
        String read2comp=new String(read2c);
        f1.append("@"+sam.getReadName()+"_1\n"+read1+"\n+\n"+qual1+"\n");
        f2.append("@"+sam.getReadName()+"_2\n"+(new StringBuilder(read2comp).reverse().toString())+"\n+\n"+(new StringBuilder(qual2).reverse().toString())+"\n");

        return 1;
    }
	

		   
	
	public void getQPosMap(int stepsize) {
	
        qposmap=new QPosMap(intervals, stepsize);
	
		if(!firstPass) {
			qposmap.readPairs(qposfileold);
			qposmap.fill(clustermap, true, qposfile,  minct1);	
		}
		else
			qposmap.fill(clustermap, false, qposfile,  minct1);	
	}
	
	 public void filterPairs(String infile, String outfile, int tolextra) {
        
        SAMFileReader tempSam=new SAMFileReader(new File(infile));
        tempSam.setValidationStringency(ValidationStringency.SILENT);
        SAMRecordIterator iter=tempSam.iterator();
        
        SAMFileWriterFactory sf2=new SAMFileWriterFactory();
        SAMFileWriter sw2=sf2.makeSAMOrBAMWriter(tempSam.getFileHeader(),false,new File(outfile));
        /*sf2.setCreateIndex(false);
        SAMFileWriter sw2=sf2.makeSAMWriter(tempSam.getFileHeader(),false,new File(outfile));*/
        
		while(iter.hasNext()) {
			SAMRecord rec=iter.next();
            if(qposmap.matchesPair(rec, intervals, tolextra)) 
                sw2.addAlignment(rec);  
        }

        sw2.close();	
        iter.close();
        tempSam.close();
    }
	
/*	public void findMates(String infile, String matefile, String outfile) {
	
		SAMFileReader tempSam=new SAMFileReader(new File(infile));
        tempSam.setValidationStringency(ValidationStringency.SILENT);
        SAMRecordIterator iter=tempSam.iterator();
		
		LinkedHashSet<String> findReads=new LinkedHashSet<String>();
   
		while(iter.hasNext()) {
			SAMRecord rec=iter.next();
			if(!rec.getReadUnmappedFlag()) {
				String name=rec.getReadName();
				String[] splname=name.split("#");
				findReads.add(splname[0]);
			}
		}
			
        iter.close();
        tempSam.close();
		
		ArrayList <String> readstofind=new ArrayList<String>(findReads);
		
		tempSam=new SAMFileReader(new File(matefile));
        tempSam.setValidationStringency(ValidationStringency.SILENT);
        iter=tempSam.iterator();
		
		SAMFileWriterFactory sf1=new SAMFileWriterFactory();
        sf1.setCreateIndex(false);
        SAMFileWriter sw1=sf1.makeSAMWriter(tempSam.getFileHeader(),false,new File(outfile));
		
		while(iter.hasNext()) {
		
			SAMRecord rec=iter.next();
						
			String nm=rec.getReadName();
			String[] splname=nm.split("#");
			String name=splname[0];
			for(int m=0; m<readstofind.size(); m++) {
				if(name.equals(readstofind.get(m))) 
					sw1.addAlignment(rec);
			}
		}
		iter.close();
		tempSam.close();
		sw1.close();

	}*/

	
	public void dumpReads(int stepsize, int tol) {
        
	
	    filterPairs(outdir+"/temp/2ends.1.bam", outdir+"/temp/2ends.filtered.2.bam", 2);
	    filterSingleEnds(outdir+"/temp/1end.filtered.1.bam", outdir+"/temp/final.singleendnm.bam", 2, false);
	    filterSingleEnds(outdir+"/temp/matenm.filtered.1.bam", outdir+"/temp/singleendmatenm.bam", 2, false);
		//findMates(outdir+"/temp/singleendmatenm.sam", outdir+"/temp/final.1.sam", outdir+"/temp/final.singleendmatenm.sam");

		List<SAMFileReader> readers=new ArrayList<SAMFileReader>();
		List<SAMFileHeader> headers=new ArrayList<SAMFileHeader>();
		SAMSequenceDictionary dict=null;
		List<File> files=new ArrayList<File>();
		files.add( new File(outdir+"/temp/1end.filtered.1.bam"));
		files.add(new File(outdir+"/temp/2ends.filtered.2.bam"));
		files.add(new File(outdir+"/temp/singleendmatenm.bam"));
		
		for(File infile:files) {
		    IoUtil.assertFileIsReadable(infile);
		    SAMFileReader in =new SAMFileReader(infile);
		    readers.add(in);
		    headers.add(in.getFileHeader());
		    if(dict==null) 
                dict=in.getFileHeader().getSequenceDictionary();
		    else if (dict.equals(in.getFileHeader().getSequenceDictionary())) 
                in.getFileHeader().setSequenceDictionary(dict);
		}
		
		File outfile=new File(outdir+"/temp/merged.sorted.bam");
        
		IoUtil.assertFileIsWritable(outfile);
		SAMFileHeader.SortOrder headerMergerSortOrder=SAMFileHeader.SortOrder.unsorted;
		SamFileHeaderMerger headerMerger=new SamFileHeaderMerger(headerMergerSortOrder, headers, true);
		MergingSamRecordIterator iterator=new MergingSamRecordIterator(headerMerger, readers, false);
		SAMFileHeader header=headerMerger.getMergedHeader();
		header.setSortOrder(SAMFileHeader.SortOrder.queryname);
		SAMRecordQueryNameComparator comp=(SAMRecordQueryNameComparator)SAMFileHeader.SortOrder.queryname.getComparatorInstance();
		SAMFileWriterFactory samFileWriterFactory=new SAMFileWriterFactory();
		SAMFileWriter out=samFileWriterFactory.makeSAMOrBAMWriter(header, false, outfile);
        
		while(iterator.hasNext()) {
		    out.addAlignment(iterator.next());
		}
		out.close();
        for(SAMFileReader reader : readers)
            reader.close();
        
		SAMFileReader tempSam=new SAMFileReader(outfile);
		tempSam.setValidationStringency(ValidationStringency.SILENT);
		SAMRecordIterator iter=tempSam.iterator();
		SAMRecord oldrec=null;
		if(iter.hasNext())
		    oldrec=iter.next();
		out=samFileWriterFactory.makeSAMOrBAMWriter(tempSam.getFileHeader(), false, new File(outdir+"/temp/merged.sorted.unique.bam"));
		String oldpair="old";
		while(iter.hasNext()) {
		    SAMRecord record=iter.next(); 
		    if(comp.compare(record, oldrec)!=0) {
                if(record.getReadName().compareTo(oldrec.getReadName())==0) {
                    if(record.getReadName().compareTo(oldpair)!=0) {
                        out.addAlignment(oldrec);
                        out.addAlignment(record);
                        oldpair=record.getReadName();
                    }
                    if(iterator.hasNext())
                        oldrec=iter.next();

                } else {
                    oldrec=record;
                }
            } else {
                oldrec=record;
		    }
		}
		out.close();
        tempSam.close();
        
        readers=new ArrayList<SAMFileReader>();
		headers=new ArrayList<SAMFileHeader>();
		dict=null;
		files=new ArrayList<File>();
		files.add( new File(outdir+"/temp/merged.sorted.unique.bam"));
		files.add(new File(outdir+"/temp/final.singleendnm.bam"));
		
		for(File infile:files) {
		    IoUtil.assertFileIsReadable(infile);
		    SAMFileReader in =new SAMFileReader(infile);
		    readers.add(in);
		    headers.add(in.getFileHeader());
		    if(dict==null) 
                dict=in.getFileHeader().getSequenceDictionary();
		    else if (dict.equals(in.getFileHeader().getSequenceDictionary())) 
                in.getFileHeader().setSequenceDictionary(dict);
		}
		
		outfile=new File(outdir+"/temp/merged.sorted.final.temp.bam");
        
        IoUtil.assertFileIsWritable(outfile);
		//SAMFileHeader.SortOrder headerMergerSortOrder=SAMFileHeader.SortOrder.unsorted;
        headerMerger=new SamFileHeaderMerger(headerMergerSortOrder, headers, true);
		iterator=new MergingSamRecordIterator(headerMerger, readers, false);
		header=headerMerger.getMergedHeader();
		header.setSortOrder(SAMFileHeader.SortOrder.queryname);
        comp=(SAMRecordQueryNameComparator)SAMFileHeader.SortOrder.queryname.getComparatorInstance();
		samFileWriterFactory=new SAMFileWriterFactory();
		out=samFileWriterFactory.makeSAMOrBAMWriter(header, false, outfile);
        
        while(iterator.hasNext()) {
		    out.addAlignment(iterator.next());
		}
		out.close();
        for(SAMFileReader reader : readers)
            reader.close();
        
        
               
        
        
	}
		    
	 	
	public void getClusterMap(int stepsize, int tol) {
		clustermap=new ClusterMap(sams, samrecs, stepsize, tol, minct1, materecs);
	}
	
	public void printClusterMap() {
		clustermap.print();
	}
	
        
    
}



